from __future__ import annotations

from advanced_alchemy.repository import SQLAlchemyAsyncRepository
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService
from ..models import RoleModel

__all__ = ("RoleService",)


class RoleService(SQLAlchemyAsyncRepositoryService[RoleModel]):
    class Repository(SQLAlchemyAsyncRepository[RoleModel]):
        model_type = RoleModel

    repository_type = Repository



from advanced_alchemy.base import UUIDAuditBase
from sqlalchemy.orm import Mapped, mapped_column, relationship


class UserModel(UUIDAuditBase):
    __tablename__ = "users"

    email: Mapped[str] = mapped_column(unique=True, nullable=False)
    password: Mapped[str] = mapped_column()

    roles = relationship(
        "RoleModel",
        secondary="users_roles",
        back_populates="users",
        lazy="selectin",
    )



from sqlalchemy.ext.asyncio.session import AsyncSession
from ..models import UserModel


async def test_create_user(db_session: AsyncSession, user_service):
    user_data = UserModel(
        email="some3@email.com",
        password="super_password",
    )
    user_data = await user_service.create(user_data, auto_commit=True)
    user = await user_service.get(user_data.id)
    assert user.email == user_data.email
    await user_service.delete(user_data.id)



async def test_get_users(db_session, users, user_service):
    users = await user_service.list()
    assert len(users) == 500





from typing import Annotated
from uuid import UUID

from advanced_alchemy.extensions.litestar.dto import SQLAlchemyDTO
from advanced_alchemy.extensions.litestar.providers import create_service_dependencies
from advanced_alchemy.filters import (
    FilterTypes,
)
from advanced_alchemy.service import OffsetPagination

from litestar import get
from litestar.controller import Controller
from litestar.dto import DTOConfig
from litestar.params import Dependency

from ..services import RoleService
from ..models import RoleModel



class RoleDTO(SQLAlchemyDTO[RoleModel]):
    config = DTOConfig(
        exclude={
            "users",
            "created_at",
            "updated_at",
            "permissions.0.created_at",
            "permissions.0.updated_at",
        },
        max_nested_depth=1,
    )


class RoleController(Controller):
    dependencies = create_service_dependencies(
        RoleService,
        key="roles_service",
        load=[RoleModel.permissions],
        filters={
            "id_filter": UUID,
            "created_at": True,
            "updated_at": True,
            "pagination_type": "limit_offset",
        },
    )

    return_dto = RoleDTO

    @get(operation_id="ListRoles", path="/roles")
    async def list_roles(
        self,
        roles_service: RoleService,
        filters: Annotated[list[FilterTypes], Dependency(skip_validation=True)],
    ) -> OffsetPagination[RoleModel]:
        results, total = await roles_service.list_and_count(*filters)
        return roles_service.to_schema(data=results, total=total, filters=filters)



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import unittest


class TestSimple(unittest.TestCase):

    def test_add_one(self):
        self.assertEqual(6, 6)


if __name__ == '__main__':
    unittest.main()



import os

import yaml
from faker import Faker
import click
from click import Group
from litestar.plugins import CLIPluginProtocol
from sqlalchemy import delete

from ..core.cli import coro
from ..core.database import session_maker
from .auth_core import hash_password
from .models import PermissionModel, RoleModel, UserModel, RolePermissionAssociation, UserRoleAssociation
from .repositories.permission_repository import provide_permission_repository
from .repositories.role_repository import provide_role_repository
from .repositories.user_repository import provide_user_repository


fake = Faker()

class AuthLoader:
    def __init__(self, session):
        self.session = session

    async def generate_users(self):
        password = hash_password("password")

        user_repository = await provide_user_repository(self.session)
        await user_repository.add_many(
            [UserModel(email=fake.unique.email(), password=password) for _ in range(497)],
            auto_commit=True
        )

    async def generate_data(self):
        permission_repository = await provide_permission_repository(self.session)
        role_repository = await provide_role_repository(self.session)
        user_repository = await provide_user_repository(self.session)

        await self.session.execute(delete(RolePermissionAssociation))
        await self.session.execute(delete(UserRoleAssociation))

        await permission_repository.delete_where()
        await role_repository.delete_where()
        await user_repository.delete_where()

        await permission_repository.delete_where(auto_commit=True)
        await role_repository.delete_where(auto_commit=True)
        await user_repository.delete_where(auto_commit=True)

        with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "./fixtures/auth.yaml"), "r") as stream:
            data_loaded = yaml.safe_load(stream)
            permissions = await permission_repository.add_many(
                [
                    PermissionModel(app=app, name=name)
                    for app, names in data_loaded.get("permissions").items()
                    for name in names
                ],
                auto_commit=True,
            )

            roles = []
            for name, permission_values in data_loaded.get("roles").items():
                role = RoleModel(name=name)
                role_permission_filter = {
                    app: {name for name in names}
                    for app, names in permission_values.items()
                }
                for permission in permissions:
                    if permission.name in role_permission_filter.get(permission.app):
                        role.permissions.append(permission)
                roles.append(role)

            await role_repository.add_many(roles, auto_commit=True)
            users = []

            for user_data in data_loaded.get("users"):
                email = user_data.get("email")
                users_roles = {role for role in user_data.get("roles")}

                user = UserModel(email=email, password=hash_password("password"))
                for role in roles:
                    if role.name in users_roles:
                        user.roles.append(role)
                users.append(user)
            await user_repository.add_many(users, auto_commit=True)

        click.secho("Success loaded auth data", fg="green")

        await self.generate_users()


class AuthCLIPlugin(CLIPluginProtocol):
    def on_cli_init(self, cli: Group) -> None:
        @cli.group(help="Manage auth, load data with ``load`` command")
        @click.version_option(prog_name='mycli')
        def auth():
            ...

        @auth.command(help="Load auth initial data")
        @coro
        async def load():
            async with session_maker() as session:
                auth_loader = AuthLoader(session)
                await auth_loader.generate_data()


from app.main import app
from collections.abc import AsyncIterator

import pytest

from litestar import Litestar
from litestar.testing import AsyncTestClient

from ..api import AccountController
from ..api.account_controller import UserDetailResponse
from ..services import UserService
from ..auth_core import encode_jwt_token
from ..models import UserModel
from ..middleware import API_KEY_HEADER, TOKEN_PREFIX


async def test_get_users(test_client: AsyncTestClient[Litestar], user_service: UserService, users) -> None:
    admin = await user_service.repository.get_one(UserModel.email=="admin@mail.com")
    user = UserDetailResponse.model_validate(admin)
    access_token = encode_jwt_token(user.model_dump(mode="json"))
    response = await test_client.get("/api/auth/users", headers={
        "Accept": "application/json",
        API_KEY_HEADER: f"{TOKEN_PREFIX}{access_token}"
    })
    response_data = response.json()
    assert len(response_data["items"]) == 20
    assert response_data['total'] == 500


async def test_login(test_client: AsyncTestClient[Litestar], user_service: UserService, users) -> None:
    admin = await user_service.repository.get_one(UserModel.email=="admin@mail.com")

    response = await test_client.post("/api/auth/account/login", json={
        "email": admin.email,
        "password": "password",
    })

async def test_get_account(test_client: AsyncTestClient[Litestar], user_service: UserService, users) -> None:
    admin = await user_service.repository.get_one(UserModel.email=="admin@mail.com")
    access_token = AccountController.generate_token(admin)
    response = await test_client.get("/api/auth/account/me", headers={
        "Accept": "application/json",
        API_KEY_HEADER: f"{TOKEN_PREFIX}{access_token}"
    })
    response_data = response.json()
    assert response_data.get('email') == admin.email



from advanced_alchemy.repository import SQLAlchemyAsyncRepository
from sqlalchemy.ext.asyncio import AsyncSession

from ..models import PermissionModel


class PermissionRepository(SQLAlchemyAsyncRepository):
    model_type = PermissionModel


async def provide_permission_repository(db_session: AsyncSession) -> PermissionRepository:
    return PermissionRepository(session=db_session)



import pytest
from typing import Dict

from ..auth_cli import AuthLoader
from ..services import UserService
from ..models import UserModel
from ..api import AccountController
from ..middleware import API_KEY_HEADER, TOKEN_PREFIX


@pytest.fixture(scope="session")
async def users(db_session):
    auth_loader = AuthLoader(db_session)
    await auth_loader.generate_data()


@pytest.fixture()
async def user_service(db_session) -> UserService:
    return UserService(db_session)


@pytest.fixture()
async def admin_headers(users, user_service: UserService) -> Dict[str, str]:
    admin = await user_service.repository.get_one(UserModel.email=="admin@mail.com")
    access_token = AccountController.generate_token(admin)
    return {
        "Accept": "application/json",
        API_KEY_HEADER: f"{TOKEN_PREFIX}{access_token}"
    }


from advanced_alchemy.base import UUIDAuditBase
from sqlalchemy import UniqueConstraint
from sqlalchemy.orm import Mapped, mapped_column, relationship


class PermissionModel(UUIDAuditBase):
    __tablename__ = "permissions"

    app: Mapped[str] = mapped_column()
    name: Mapped[str] = mapped_column()

    roles = relationship(
        "RoleModel",
        secondary="roles_permissions",
        back_populates="permissions",
    )

    __table_args__ = (UniqueConstraint("app", "name", name="_permission_info"),)



"""
Configure litestar app with lab module:

from litestar import Litestar

from lab.auth.api import auth_router
from lab.auth.auth_cli import CLIPlugin
from lab.core.database import alchemy
from lab.core.openapi import configure_open_api_config


app = Litestar(
    route_handlers=[index, auth_router],
    plugins=[alchemy, CLIPlugin()],
    openapi_config=configure_open_api_config("service template", "2.0.0"),
    debug=True
)
"""



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


DEFAULT_DB_URI = "postgresql+asyncpg://user:password@localhost:5432/database"

class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")
    database_uri: str = Field(default=DEFAULT_DB_URI)
    secret_key: str = Field(default="Secret_Key")


settings = Settings()



from typing import List
from uuid import UUID

import advanced_alchemy
from litestar import Request, get, post
from litestar.connection import ASGIConnection
from litestar.controller import Controller
from litestar.di import Provide
from litestar.exceptions import NotAuthorizedException, NotFoundException
from litestar.handlers.base import BaseRouteHandler
from pydantic import BaseModel, ConfigDict, SecretStr

from ..auth_core import check_password, encode_jwt_token
from ..models import UserModel
from ..repositories import UserRepository, provide_user_repository


class Base(BaseModel):
    model_config = ConfigDict(from_attributes=True)


class Role(Base):
    id: UUID | None
    name: str


class UserDetailResponse(Base):
    id: UUID | None
    email: str
    password: SecretStr
    roles: List[Role]


class LoginRequest(BaseModel):
    email: str
    password: str


class LoginResponse(BaseModel):
    access_token: str
    user: UserDetailResponse


def admin_user_guard(connection: ASGIConnection, _: BaseRouteHandler) -> None:
    if connection.user.get("role") != "admin":
        raise NotAuthorizedException()


class AccountController(Controller):
    @staticmethod
    def generate_token(user_data: UserModel) -> str:
        user = UserDetailResponse.model_validate(user_data)
        return encode_jwt_token(user.model_dump(mode="json"))

    dependencies = {
        "user_repository": Provide(provide_user_repository),
    }
    path = "/account"

    @post(
        path="/login",
        exclude_from_auth=True,
    )
    async def login(
            self,
            user_repository: UserRepository,
            data: LoginRequest,
    ) -> LoginResponse:
        try:
            admin = await user_repository.get_one(UserModel.email == data.email)
        except advanced_alchemy.exceptions.NotFoundError:
            raise NotFoundException(f"User with email {data.email} not found")

        if not check_password(
            data.password,
            admin.password,
        ):
            raise NotFoundException("Invalid password")

        access_token = self.generate_token(admin)

        return LoginResponse(
            user=UserDetailResponse.model_validate(admin),
            access_token=access_token,
        )

    @get(path="/me")
    async def me(self, request: Request) -> UserDetailResponse:
        if not request.user:
            raise NotAuthorizedException()

        return UserDetailResponse(**request.user)
